from twitter.pants.targets.jvm_target import JvmTarget

class OinkQuery(JvmTarget):

  def __init__(self, name, dependencies, sources=None, excludes=None):
    JvmTarget.__init__(self, name, sources, dependencies, excludes)
    # TODO: configurations is required when fetching jar_dependencies but should not be
    self.configurations = None



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'John Sirois'

from twitter.pants import is_python
from twitter.pants.base.builder import Builder
from twitter.pants.targets import PythonBinary, PythonTests, PythonTestSuite

from twitter.pants.python.binary_builder import PythonBinaryBuilder
from twitter.pants.python.test_builder import PythonTestBuilder
from twitter.pants.python.lint_builder import PythonLintBuilder

class PythonBuilder(Builder):
  def __init__(self, ferror, root_dir):
    Builder.__init__(self, ferror, root_dir)

  def build(self, targets, args, conn_timeout=None):
    test_targets = []
    binary_targets = []

    for target in targets:
      assert is_python(target), "PythonBuilder can only build PythonTargets, given %s" % str(target)

    if 'pylint' in args:
      real_args = list(args)
      real_args.remove('pylint')
      for target in targets:
        try:
          PythonLintBuilder([target], real_args, self.root_dir, conn_timeout=conn_timeout).run()
        except Exception as e:
          print('Failed to run lint for %s: %s' % (target, e))
      return 0

    # PythonBuilder supports PythonTests and PythonBinaries
    for target in targets:
      if isinstance(target, PythonTests) or isinstance(target, PythonTestSuite):
        test_targets.append(target)
      elif isinstance(target, PythonBinary):
        binary_targets.append(target)

    rv = PythonTestBuilder(test_targets, args, self.root_dir, conn_timeout=conn_timeout).run()
    if rv != 0: return rv

    for binary_target in binary_targets:
      rv = PythonBinaryBuilder(binary_target, args, self.root_dir, conn_timeout=conn_timeout).run()
      if rv != 0: return rv

    return 0



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import os
import subprocess
import sys

from twitter.common.dirutil import safe_rmtree
from twitter.pants import get_buildroot
from twitter.pants.python.code_generator import CodeGenerator
from twitter.pants.targets.python_thrift_library import PythonThriftLibrary
from twitter.pants.thrift_util import calculate_compile_roots, select_thrift_binary


class PythonThriftBuilder(CodeGenerator):
  class UnknownPlatformException(CodeGenerator.Error):
    def __init__(self, platform):
      super(PythonThriftBuilder.UnknownPlatformException, self).__init__(
          "Unknown platform: %s!" % str(platform))

  def run_thrifts(self):
    def is_py_thrift(target):
      return isinstance(target, PythonThriftLibrary)
    bases, roots = calculate_compile_roots([self.target], is_py_thrift)

    for src in roots:
      if not self._run_thrift(src, bases):
        raise PythonThriftBuilder.CodeGenerationException(
          "Could not generate .py from %s!" % src)

  def _run_thrift(self, source, bases):
    thrift_file = source
    thrift_abs_path = os.path.abspath(os.path.join(self.root, thrift_file))

    args = [
      select_thrift_binary(self.config),
      '--gen',
      'py:new_style',
      '-recurse',
      '-o',
      self.codegen_root
    ]

    # Add bases as include paths to try.  Note that include paths and compile targets
    # should be uniformly relative, or uniformly absolute (in this case the latter).
    for base in bases:
      args.extend(('-I', os.path.join(get_buildroot(), base)))
    args.append(thrift_abs_path)

    po = subprocess.Popen(args, cwd=self.chroot.path())
    rv = po.wait()
    if rv != 0:
      comm = po.communicate()
      print('thrift generation failed!', file=sys.stderr)
      print('STDOUT', file=sys.stderr)
      print(comm[0], file=sys.stderr)
      print('STDERR', file=sys.stderr)
      print(comm[1], file=sys.stderr)
    return rv == 0

  @property
  def package_dir(self):
    return "gen-py"

  def generate(self):
    # autogenerate the python files that we bundle up
    self.run_thrifts()

    # Thrift generates code with all parent namespaces with empty __init__.py's. Generally
    # speaking we want to drop anything w/o an __init__.py, and for anything with an __init__.py,
    # we want to explicitly make it a namespace package, hence the hoops here.
    for root, _, files in os.walk(os.path.normpath(self.package_root)):
      reldir = os.path.relpath(root, self.package_root)
      if reldir == '.':  # skip root
        continue
      if '__init__.py' not in files:  # skip non-packages
        continue
      init_py_abspath = os.path.join(root, '__init__.py')
      module_path = self.path_to_module(reldir)
      self.created_packages.add(module_path)
      if os.path.getsize(init_py_abspath) == 0:  # empty __init__, translate to namespace package
        with open(init_py_abspath, 'wb') as f:
          f.write(b"__import__('pkg_resources').declare_namespace(__name__)")
        self.created_namespace_packages.add(module_path)
      else:
        # non-empty __init__, this is a leaf package, usually with ttypes and constants, leave as-is
        pass

    if not self.created_packages:
      raise self.CodeGenerationException(
        'No Thrift structures declared in %s!' % self.target)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

__author__ = "tdesai"

from twitter.pants.goal.jvm_tool_goal import JvmToolGoal
from twitter.pants.tasks.jvm_tool_task import JvmToolTask
from twitter.pants.task_specs.arguments import JvmProperty
from twitter.pants.task_specs.nailgun_task_spec_with_jar_deps import NailgunTaskSpecWithJarDeps


class JvmToolSpec(NailgunTaskSpecWithJarDeps):

  def __init__(self,
               name,
               main,
               sources,
               args,
               dependencies,
               ng_daemon=None,
               options=None,
               description='',
               phase_dependencies=None):
    super(JvmToolSpec, self).__init__(name=name,
                                      main=main,
                                      sources=sources,
                                      args=args,
                                      dependencies=dependencies,
                                      ng_daemon=ng_daemon,
                                      options=options,
                                      description=description,
                                      phase_dependencies=phase_dependencies)
    goal = JvmToolGoal(name=name, action=JvmToolTask, jvm_tool_spec=self,
                       dependencies=phase_dependencies)
    phase = goal.install()
    if description:
      phase.with_description(description)

  def _add_default_options(self):
    skip_tool_option = {'dest': '_'.join([self._name, 'skip']),
                        'name': 'skip',
                        'default': False,
                        'help': 'Skip %s' % self._name}
    self._options.append(skip_tool_option)

  def get_args(self, config):
    args = {'jvmargs': [],
            'opts': []}
    for arg in self._args:
      key = 'jvmargs' if isinstance(arg, JvmProperty) else 'opts'
      args[key].extend(arg.get(config))
    return args



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import os
import sys

from twitter.common.collections import OrderedSet
from twitter.common.python.pex import PEX

from twitter.pants.base import Target, Address
from twitter.pants.python.python_chroot import PythonChroot

try:
  import pylint
  _HAVE_PYLINT = True
except ImportError:
  _HAVE_PYLINT = False

class PythonLintBuilder(object):
  def __init__(self, targets, args, root_dir, conn_timeout=None):
    self.targets = targets
    self.args = args
    self.root_dir = root_dir
    self._conn_timeout = conn_timeout

  def run(self):
    if not _HAVE_PYLINT:
      print('ERROR: Pylint not found!  Skipping.', file=sys.stderr)
      return 1
    return self._run_lints(self.targets, self.args)

  def _run_lint(self, target, args):
    chroot = PythonChroot(target, self.root_dir, extra_targets=[
      Target.get(Address.parse(self.root_dir, '3rdparty/python:pylint'))],
      conn_timeout=self._conn_timeout)
    chroot.builder.info.ignore_errors = True
    builder = chroot.dump()
    builder.info.entry_point = 'pylint.lint'
    builder.freeze()

    interpreter_args = [
      '--rcfile=%s' % os.path.join(self.root_dir, 'build-support', 'pylint', 'pylint.rc')]
    interpreter_args.extend(args or [])
    sources = OrderedSet([])
    target.walk(lambda trg: sources.update(
      trg.sources if hasattr(trg, 'sources') and trg.sources is not None else []))
    pex = PEX(builder.path())
    pex.run(args=interpreter_args + list(sources), with_chroot=True)

  def _run_lints(self, targets, args):
    for target in targets:
      self._run_lint(target, args)
    return 0



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import traceback

from . import Command

from twitter.common.collections import OrderedSet
from twitter.pants import is_concrete, is_python
from twitter.pants.base import Address, Config, Target
from twitter.pants.targets import InternalTarget
from twitter.pants.python import PythonBuilder

class Build(Command):
  """Builds a specified target."""

  __command__ = 'build'

  def setup_parser(self, parser, args):
    parser.set_usage("\n"
                     "  %prog build (options) [spec] (build args)\n"
                     "  %prog build (options) [spec]... -- (build args)")
    parser.add_option("-t", "--timeout", dest="conn_timeout", type="int",
                      default=Config.load().getdefault('connection_timeout'),
                      help="Number of seconds to wait for http connections.")
    parser.disable_interspersed_args()
    parser.epilog = """Builds the specified python target(s)."""

  def __init__(self, root_dir, parser, argv):
    Command.__init__(self, root_dir, parser, argv)

    if not self.args:
      self.error("A spec argument is required")

    try:
      specs_end = self.args.index('--')
      if len(self.args) > specs_end:
        self.build_args = self.args[specs_end+1:len(self.args)+1]
      else:
        self.build_args = []
    except ValueError:
      specs_end = 1
      self.build_args = self.args[1:] if len(self.args) > 1 else []

    self.targets = OrderedSet()
    for spec in self.args[0:specs_end]:
      try:
        address = Address.parse(root_dir, spec)
      except:
        self.error("Problem parsing spec %s: %s" % (spec, traceback.format_exc()))

      try:
        target = Target.get(address)
      except:
        self.error("Problem parsing BUILD target %s: %s" % (address, traceback.format_exc()))

      try:
        InternalTarget.check_cycles(target)
      except InternalTarget.CycleException as e:
        self.error("Target contains an internal dependency cycle: %s" % e)

      if not target:
        self.error("Target %s does not exist" % address)
      self.targets.update(tgt for tgt in target.resolve() if is_concrete(tgt))

  def execute(self):
    print("Build operating on targets: %s" % self.targets)

    python_targets = OrderedSet()
    for target in self.targets:
      if is_python(target):
        python_targets.add(target)
      else:
        self.error("Cannot build target %s" % target)

    if python_targets:
      status = self._python_build(python_targets)

    return status

  def _python_build(self, targets):
    try:
      executor = PythonBuilder(self.error, self.root_dir)
      return executor.build(targets, self.build_args, conn_timeout=self.options.conn_timeout)
    except:
      self.error("Problem executing PythonBuilder for targets %s: %s" % (targets,
                                                                         traceback.format_exc()))



from __future__ import print_function

from twitter.common.collections import OrderedSet
from twitter.common.python.fetcher import Fetcher, PyPIFetcher
from twitter.common.python.http import Crawler
from twitter.common.python.obtainer import Obtainer
from twitter.common.python.platforms import Platform
from twitter.common.python.resolver import ResolverBase
from twitter.common.python.translator import (
    ChainedTranslator,
    EggTranslator,
    SourceTranslator)

from twitter.pants.targets import (
    PythonBinary,
    PythonRequirement)


def get_platforms(platform_list):
  def translate(platform):
    return Platform.current() if platform == 'current' else platform
  return tuple(map(translate, platform_list))


class MultiResolver(ResolverBase):
  """A multi-platform PythonRequirement resolver for Pants."""

  @classmethod
  def fetchers(cls, config, target=None):
    fetchers = []
    fetchers.extend(Fetcher([url]) for url in config.getlist('python-repos', 'repos', []))
    fetchers.extend(PyPIFetcher(url) for url in config.getlist('python-repos', 'indices', []))
    if target and isinstance(target, PythonBinary):
      fetchers.extend(Fetcher([url]) for url in target.repositories)
      fetchers.extend(PyPIFetcher(url) for url in target.indices)
    return fetchers

  @classmethod
  def crawler(cls, config, conn_timeout=None):
    return Crawler(cache=config.get('python-setup', 'download_cache'),
                   conn_timeout=conn_timeout)

  def __init__(self, config, target, conn_timeout=None):
    platforms = config.getlist('python-setup', 'platforms', ['current'])
    if isinstance(target, PythonBinary) and target.platforms:
      platforms = target.platforms
    self._install_cache = config.get('python-setup', 'install_cache')
    self._crawler = self.crawler(config, conn_timeout=conn_timeout)
    self._fetchers = self.fetchers(config, target)
    self._platforms = get_platforms(platforms)
    super(MultiResolver, self).__init__(cache=self._install_cache)

  def make_installer(self, reqs, interpreter, platform):
    assert len(reqs) == 1 and isinstance(reqs[0], PythonRequirement), 'Got requirement list: %s' % (
      repr(reqs))
    req = reqs[0]
    fetchers = [Fetcher([req.repository])] + self._fetchers if req.repository else self._fetchers
    translator = ChainedTranslator(
      EggTranslator(install_cache=self._install_cache, platform=platform,
          python=interpreter.python),
      SourceTranslator(install_cache=self._install_cache, interpreter=interpreter,
          platform=platform, use_2to3=req.use_2to3))
    obtainer = Obtainer(self._crawler, fetchers, translator)
    return obtainer.obtain

  def resolve(self, requirements, interpreter=None):
    resolved = OrderedSet()
    requirements = list(requirements)
    for platform in self._platforms:
      for req in requirements:
        resolved.update(super(MultiResolver, self).resolve(req, platform=platform,
            interpreter=interpreter))
    return list(resolved)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================
__author__ = "tdesai"

from twitter.common.collections import maybe_list
from twitter.pants.targets.jar_dependency import JarDependency
from twitter.pants.task_specs import TaskSpec, TaskSpecError


class NailgunTaskSpecWithJarDeps(TaskSpec):
  def __init__(self,
               name,
               main,
               sources,
               args,
               dependencies,
               ng_daemon=None,
               options=None,
               description='',
               phase_dependencies=None):
    super(NailgunTaskSpecWithJarDeps, self).__init__(name=name,
                                                     sources=sources,
                                                     args=args,
                                                     options=options,
                                                     phase_dependencies=phase_dependencies,
                                                     description=description)
    self._main = main
    self._ng_daemon = ng_daemon
    self._dependencies = maybe_list(dependencies, JarDependency, TaskSpecError)

  @property
  def dependencies(self):
    return self._dependencies

  @property
  def main(self):
    return self._main

  @property
  def ng_daemon(self):
    return self._ng_daemon



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

__author__ = "tdesai"

import os
import pkgutil

from twitter.common.dirutil import safe_mkdir, touch
from twitter.pants import binary_util, get_buildroot
from twitter.pants.base.generator import Generator, TemplateData
from twitter.pants.tasks import TaskError


DEFAULT_CONFS = ['default']
TEMPLATE_PATH = os.path.join('templates', 'ivy_resolve', 'ivy.mk')


#TODO: Try to refactor ivy_resolve.py and use that
#Refer AWESOME-4329
class IvyResolver(object):
  def __init__(self, context, jvm_tool_spec):
    self._config = context.config
    self._jvm_tool_spec = jvm_tool_spec
    self._name = self._jvm_tool_spec.name
    self._dependencies = self._jvm_tool_spec.dependencies

  def make_runnable(self):
    self._workdir = self._config.get(self._name, 'workdir',
                                     default=os.path.join(self._config.getdefault('homedir'),
                                                          '.pants.d',
                                                          self._name))
    safe_mkdir(self._workdir)
    self._confs = self._config.getlist(self._name, 'confs', default=DEFAULT_CONFS)
    ivy_xml = self._generate_ivy_xml()
    self._run_ivy_resolve(ivy_xml)

  def _generate_ivy_xml(self):
    ivy_xml = os.path.join(self._workdir, 'ivy.xml')
    template_data = TemplateData(
        org="com.twitter",
        module=self._name,
        version='',
        publications=None,
        is_idl=False,
        dependencies=[self._generate_jar_template(jar) for jar in self._dependencies],
        excludes=[]
    )
    with open(ivy_xml, 'w') as output:
      generator = Generator(pkgutil.get_data(__name__, TEMPLATE_PATH),
                            lib=template_data)
      generator.write(output)
    return ivy_xml

  def _generate_jar_template(self, jar):
    return TemplateData(org=jar.org,
                        module=jar.name,
                        version=jar.rev,
                        mutable=self._is_mutable(jar),
                        force=jar.force,
                        excludes=[self._generate_exclude_template(exclude)
                                    for exclude in jar.excludes],
                        transitive=jar.transitive,
                        artifacts=jar.artifacts,
                        is_idl='idl' in jar._configurations,
                        configurations=';'.join(jar._configurations))

  def _is_mutable(self, jar):
    if jar.mutable is not None:
      return jar.mutable
            #TODO: setup_parser
            #if self._mutable_pattern:
            #return self._mutable_pattern.match(jar.rev)
    return False

  def _generate_exclude_template(self, exclude):
    return TemplateData(org=exclude.org, name=exclude.name)

  def _run_ivy_resolve(self, ivy_xml, ivy_settings=None):
    #TODO: https://jira.twitter.biz/browse/AWESOME-4330
    tool_libdir = os.path.join(self._workdir, 'lib')
    tool_check = '%s.checked' % tool_libdir
    if not os.path.exists(tool_check):
      safe_mkdir(tool_libdir)
      ivy_settings = ivy_settings or self._config.get('ivy', 'ivy_settings')
      ivy_opts = ['-settings', ivy_settings,
                  '-ivy', ivy_xml,
                  '-cache', self._config.get('ivy-resolve', 'cache_dir'),
                  '-retrieve', '%s/[artifact]-[revision](-[classifier]).[ext]' % tool_libdir,
                  '-confs', ','.join(self._confs)]
      classpath = binary_util.get_tool_classpath(tool_lib_dir=os.path.join(get_buildroot(),
                                                                           'build-support',
                                                                           'ivy',
                                                                           'lib'))
      result = binary_util.runjava_indivisible(main='org.apache.ivy.Main',
                                               classpath=classpath,
                                               opts=ivy_opts)
      if result != 0:
        raise TaskError('Failed to load tool %s, ivy exit code %d' % (self._name, result))
      touch(tool_check)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from twitter.pants.base import manual
from twitter.pants.targets.exclude import Exclude

from .external_dependency import ExternalDependency


class Artifact(object):
  """
  Specification for an Ivy Artifact for this jar dependency.

  See: http://ant.apache.org/ivy/history/latest-milestone/ivyfile/artifact.html
  """
  def __init__(self, name, type_=None, ext=None, conf=None, url=None, classifier=None):
    """
    :param name: The name of the published artifact. This name must not include revision.
    :param type_: The type of the published artifact. It's usually its extension, but not
      necessarily. For instance, ivy files are of type 'ivy' but have 'xml' extension.
    :param ext: The extension of the published artifact.
    :param conf: The public configuration in which this artifact is published. The '*' wildcard can
      be used to designate all public configurations.
    :param url: The url at which this artifact can be found if it isn't located at the standard
      location in the repository
    :param classifier: The maven classifier of this artifact.
    """
    self.name = name
    self.type_ = type_ or 'jar'
    self.ext = ext
    self.url = url
    self.classifier = classifier
    self.conf = conf

  def __repr__(self):
    return ('Artifact(%r, type_=%r, ext=%r, conf=%r, url=%r, classifier=%r)'
            % (self.name, self.type_, self.ext, self.conf, self.url, self.classifier))


@manual.builddict(tags=["jvm"])
class  JarDependency(ExternalDependency):
  """Represents a binary jar dependency ala maven or ivy.  For the ivy dependency defined by:
  ..

    <dependency org="com.google.guava" name="guava" rev="r07"/>

  The equivalent Dependency object could be created with:
    JarDependency(org = "com.google.guava", name = "guava", rev = "r07")

  If the rev keyword argument is left out, the revision is assumed to be the latest available.

  If the rev is specified and force = True is also specified, this will force the artifact revision
  to be rev even if other transitive deps specify a different revision for the same artifact.

  The extension of the artifact can be over-ridden if it differs from the artifact type with the ext
  keyword argument.  This is sometimes needed for artifacts packaged with maven bundle type but
  stored as jars.

  The url of the artifact can be over-ridden if non-standard by specifying the url keyword argument.

  If the dependency has API docs available online, these can be noted with apidocs and generated
  javadocs with {@link}s to the jar's classes will be properly hyperlinked.

  If the dependency is mutable this must be explicitly noted.  A common use-case is to inhibit
  caching of maven -SNAPSHOT style artifacts in an active development/integration cycle.

  If you want to use a maven classifier variant of a jar, use the classifier param. If you want
  to include multiple artifacts with differing classifiers, use with_artifact.
  """
  _JAR_HASH_KEYS = (
    'org',
    'name',
    'rev',
    'force',
    'excludes',
    'transitive',
    'ext',
    'url',
    'mutable',
    '_configurations'
  )

  def __init__(self, org, name, rev=None, force=False, ext=None, url=None, apidocs=None,
               type_=None, classifier=None, mutable=None):
    self.org = org
    self.name = name
    self.rev = rev
    self.force = force
    self.excludes = []
    self.transitive = True
    self.apidocs = apidocs
    self.mutable = mutable
    self._classifier = classifier

    self.artifacts = []
    if ext or url or type_ or classifier:
      self.with_artifact(name=name, type_=type_, ext=ext, url=url, classifier=classifier)

    self.id = repr(self)
    self._configurations = ['default']

    # Support legacy method names
    # TODO(John Sirois): introduce a deprecation cycle for these and then kill
    self.withSources = self.with_sources
    self.withDocs = self.with_sources

  @property
  def classifier(self):
    """Returns the maven classifier for this jar dependency.

    If the classifier is ambiguous; ie: there was no classifier set in the constructor and the jar
    dependency has multiple attached artifacts, a :class:`ValueError` is raised.
    """
    if self._classifier or len(self.artifacts) == 0:
      return self._classifier
    elif len(self.artifacts) == 1:
      return self.artifacts[0].classifier
    else:
      raise ValueError('Cannot determine classifier. No explicit classifier is set and this jar '
                       'has more than 1 artifact: %s\n\t%s'
                       % (self, '\n\t'.join(map(str, self.artifacts))))

  @manual.builddict()
  def exclude(self, org, name=None):
    """Adds a transitive dependency of this jar to the exclude list."""

    self.excludes.append(Exclude(org, name))
    return self

  @manual.builddict()
  def intransitive(self):
    """Declares this Dependency intransitive, indicating only the jar for the dependency itself
    should be downloaded and placed on the classpath"""

    self.transitive = False
    return self

  @manual.builddict()
  def with_sources(self):
    """This requests the artifact have its source jar fetched.
    (This implies there *is* a source jar to fetch.) Used in contexts
    that can use source jars (as of 2013, just eclipse and idea goals)."""
    self._configurations.append('sources')
    return self

  def with_docs(self):
    self._configurations.append('docs')
    return self

  def with_artifact(self, name=None, type_=None, ext=None, url=None, configuration=None,
                    classifier=None):
    """Sets an alternative artifact to fetch or adds additional artifacts if called multiple times.
    """
    artifact = Artifact(name or self.name, type_=type_, ext=ext, url=url, conf=configuration,
                        classifier=classifier)
    self.artifacts.append(artifact)
    return self

  def __eq__(self, other):
    result = (isinstance(other, JarDependency)
              and self.org == other.org
              and self.name == other.name
              and self.rev == other.rev)
    return result

  def __hash__(self):
    return hash((self.org, self.name, self.rev))

  def __ne__(self, other):
    return not self.__eq__(other)

  def __repr__(self):
    return "%s-%s-%s" % (self.org, self.name, self.rev)

  def cache_key(self):
    return ''.join(getattr(self, key) for key in self._JAR_HASH_KEYS)

  def resolve(self):
    yield self

  def walk(self, work, predicate=None):
    if not predicate or predicate(self):
      work(self)

  def _as_jar_dependencies(self):
    yield self



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.pants import binary_util, is_codegen, is_java
from twitter.pants.task_specs.arguments import ArgError
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.nailgun_task import NailgunTask


DEFAULT_CONFS = ['default']


def _is_checked(target):
  return is_java(target) and not is_codegen(target)


class JvmToolTask(NailgunTask):

  @classmethod
  def setup_parser(cls, option_group, args, mkflag, options):
    NailgunTask.setup_parser(option_group, args, mkflag)
    for option in options:
      option_group.add_option(mkflag(option['name']),
                              mkflag(option['name'], negate=True),
                              dest=option['dest'],
                              default=option['default'],
                              action='callback', callback=mkflag.set_bool,
                              help=option['help'])

  def __init__(self, context, jvm_tool_spec):
    self._jvm_tool_spec = jvm_tool_spec
    self._name = jvm_tool_spec.name
    self.__class__.__name__ = self._name
    self._workdir = context.config.get(
                      self._name, 'workdir',
                      default=os.path.join(context.config.getdefault('homedir'),
                                           '.pants.d',
                                           self._name))
    self._confs = context.config.getlist(self._name, 'confs', default=DEFAULT_CONFS)
    self._log = context.log
    super(JvmToolTask, self).__init__(context, use_daemon=jvm_tool_spec.ng_daemon)

  def execute(self, targets):
    try:
      if not getattr(self.context.options, '_'.join((self._name, "skip"))):
        with self.changed(filter(_is_checked, targets)) as changed_targets:
          sources = self._calculate_sources(changed_targets)
          if sources:
            result = self._run_tool(sources)
            if result != 0:
              raise TaskError('java %s ... exited non-zero (%i)' % (self._jvm_tool_spec.main,
                                                                    result))
    except ArgError as ae:
      self._log.info("%s" % ae)
      raise TaskError('Arguments specified to tool are incorrect.\n%s' % ae)

  def _calculate_sources(self, targets):
    sources = set()
    for target in targets:
      for suffix in self._jvm_tool_spec.sources:
        sources.update([os.path.join(target.target_base, source)
                        for source in target.sources if source.endswith(suffix)])
    return sources

  def _run_tool(self, sources):
    classpath = binary_util.get_tool_classpath(tool_lib_dir=os.path.join(self._workdir, 'lib'))
    with self.context.state('classpath', []) as cp:
      classpath.extend(jar for conf, jar in cp if conf in self._confs)
    args = self._jvm_tool_spec.get_args(self.context.config)
    return self.runjava(main=self._jvm_tool_spec.main,
                        classpath=classpath,
                        opts=args['opts'],
                        jvmargs=args['jvmargs'],
                        args=sources)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

from collections import defaultdict
import os
import random
import sys
import tempfile

from twitter.common.collections import OrderedSet
from twitter.common.decorators import lru_cache
from twitter.common.dirutil import safe_rmtree
from twitter.common.python.interpreter import PythonInterpreter
from twitter.common.python.pex_builder import PEXBuilder
from twitter.common.python.platforms import Platform
from twitter.common.quantity import Amount, Time
from twitter.pants import is_concrete
from twitter.pants.base import Config, ParseContext, TargetDefinitionException
from twitter.pants.base.build_cache import BuildCache
from twitter.pants.targets import (
    PythonAntlrLibrary,
    PythonBinary,
    PythonLibrary,
    PythonRequirement,
    PythonTests,
    PythonThriftLibrary)

from .antlr_builder import PythonAntlrBuilder
from .resolver import MultiResolver
from .thrift_builder import PythonThriftBuilder


class PythonChroot(object):
  _VALID_DEPENDENCIES = {
    PythonLibrary: 'libraries',
    PythonRequirement: 'reqs',
    PythonBinary: 'binaries',
    PythonThriftLibrary: 'thrifts',
    PythonAntlrLibrary: 'antlrs',
    PythonTests: 'tests'
  }

  MEMOIZED_THRIFTS = {}

  class InvalidDependencyException(Exception):
    def __init__(self, target):
      Exception.__init__(self, "Not a valid Python dependency! Found: %s" % target)

  def __init__(self, target, root_dir, extra_targets=None, builder=None, interpreter=None,
      conn_timeout=None):
    self._config = Config.load()
    self._target = target
    self._root = root_dir
    self._interpreter = interpreter or PythonInterpreter.get()
    self._cache = BuildCache(os.path.join(self._config.get('python-setup', 'artifact_cache'),
      '%s' % self._interpreter.identity))
    self._extra_targets = list(extra_targets) if extra_targets is not None else []
    self._resolver = MultiResolver(self._config, target, conn_timeout=conn_timeout)
    self._builder = builder or PEXBuilder(tempfile.mkdtemp(), interpreter=self._interpreter)

  def __del__(self):
    if os.getenv('PANTS_LEAVE_CHROOT') is None:
      safe_rmtree(self.path())

  @property
  def builder(self):
    return self._builder

  def debug(self, msg, indent=0):
    if os.getenv('PANTS_VERBOSE') is not None:
      print('%s%s' % (' ' * indent, msg))

  def path(self):
    return self._builder.path()

  def _dump_library(self, library):
    def copy_to_chroot(base, path, add_function):
      src = os.path.join(self._root, base, path)
      add_function(src, path)

    self.debug('  Dumping library: %s' % library)
    for filename in library.sources:
      copy_to_chroot(library.target_base, filename, self._builder.add_source)
    for filename in library.resources:
      copy_to_chroot(library.target_base, filename, self._builder.add_resource)

  def _dump_requirement(self, req, dynamic, repo):
    self.debug('  Dumping requirement: %s%s%s' % (str(req),
      ' (dynamic)' if dynamic else '', ' (repo: %s)' % repo if repo else ''))
    self._builder.add_requirement(req, dynamic, repo)

  def _dump_distribution(self, dist):
    self.debug('  Dumping distribution: .../%s' % os.path.basename(dist.location))
    self._builder.add_distribution(dist)

  def _generate_requirement(self, library, builder_cls):
    library_key = self._cache.key_for_target(library)
    builder = builder_cls(library, self._root, self._config, '-' + library_key.hash[:8])
    if self._cache.needs_update(library_key):
      sdist = builder.build(interpreter=self._interpreter)
      self._cache.update(library_key, build_artifacts=[sdist], artifact_root=os.path.dirname(sdist))
    with ParseContext.temp():
      return PythonRequirement(builder.requirement_string(), repository=library_key.filename,
        use_2to3=True)

  def _generate_thrift_requirement(self, library):
    return self._generate_requirement(library, PythonThriftBuilder)

  def _generate_antlr_requirement(self, library):
    return self._generate_requirement(library, PythonAntlrBuilder)

  def resolve(self, targets):
    children = defaultdict(OrderedSet)
    def add_dep(trg):
      if is_concrete(trg):
        for target_type, target_key in self._VALID_DEPENDENCIES.items():
          if isinstance(trg, target_type):
            children[target_key].add(trg)
            return
      raise self.InvalidDependencyException(trg)
    for target in targets:
      target.walk(add_dep)
    return children

  def dump(self):
    self.debug('Building PythonBinary %s:' % self._target)

    targets = self.resolve([self._target] + self._extra_targets)

    for lib in targets['libraries'] | targets['binaries']:
      self._dump_library(lib)

    generated_reqs = OrderedSet()
    if targets['thrifts']:
      for thr in set(targets['thrifts']):
        if thr not in self.MEMOIZED_THRIFTS:
          self.MEMOIZED_THRIFTS[thr] = self._generate_thrift_requirement(thr)
        generated_reqs.add(self.MEMOIZED_THRIFTS[thr])
      with ParseContext.temp():
        # trick pants into letting us add this python requirement, otherwise we get
        # TargetDefinitionException: Error in target BUILD.temp:thrift: duplicate to
        # PythonRequirement(thrift)
        #
        # TODO(wickman) Instead of just blindly adding a PythonRequirement for thrift, we
        # should first detect if any explicit thrift requirements have been added and use
        # those.  Only if they have not been supplied should we auto-inject it.
        generated_reqs.add(PythonRequirement('thrift', use_2to3=True,
            name='thrift-' + ''.join(random.sample('0123456789abcdef' * 8, 8))))

    for antlr in targets['antlrs']:
      generated_reqs.add(self._generate_antlr_requirement(antlr))

    targets['reqs'] |= generated_reqs
    for req in targets['reqs']:
      if not req.should_build(self._interpreter.python, Platform.current()):
        self.debug('Skipping %s based upon version filter' % req)
        continue
      self._dump_requirement(req._requirement, False, req._repository)

    reqs_to_build = (req for req in targets['reqs']
        if req.should_build(self._interpreter.python, Platform.current()))
    for dist in self._resolver.resolve(reqs_to_build, interpreter=self._interpreter):
      self._dump_distribution(dist)

    if len(targets['binaries']) > 1:
      print('WARNING: Target has multiple python_binary targets!', file=sys.stderr)

    return self._builder



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'Tejal Desai'

from .stats_http_client import StatsHttpClient, StatsUploader
from .buildtime_stats import BuildTimeStats
__all__ = (
  'StatsHttpClient',
  'StatsUploader'
  'BuildTimeStats',
)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

__author__ = 'Brian Wickman'

import os
import tempfile

from twitter.common.python.pex_builder import PEXBuilder
from twitter.pants.base import Config
from twitter.pants.targets import PythonBinary
from twitter.pants.python.python_chroot import PythonChroot


class PythonBinaryBuilder(object):
  class NotABinaryTargetException(Exception): pass

  def __init__(self, target, args, root_dir, conn_timeout=None):
    self.target = target
    if not isinstance(target, PythonBinary):
      raise PythonBinaryBuilder.NotABinaryTargetException(
        "Target %s is not a PythonBinary!" % target)
    config = Config.load()
    self.distdir = config.getdefault('pants_distdir')
    distpath = tempfile.mktemp(dir=self.distdir, prefix=target.name)
    self.builder = PEXBuilder(distpath, pex_info=target.pexinfo)
    self.chroot = PythonChroot(target, root_dir, builder=self.builder, conn_timeout=conn_timeout)

  def run(self):
    print('Building PythonBinary %s:' % self.target)
    env = self.chroot.dump()
    filename = os.path.join(self.distdir, '%s.pex' % self.target.name)
    env.build(filename)
    print('Wrote %s' % filename)
    return 0



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = "tdesai"

import itertools


class TaskSpecError(Exception):
  """Raised to indicate a given TaskSpec is not valid."""


class TaskSpec(object):

  def __init__(self,
               name,
               sources,
               args,
               options=None,
               description='',
               phase_dependencies=None):
    self._name = name
    self._sources = sources
    self._args = args
    #Todo: For now options is an array of hashes with keys 'dest', 'help', 'default'
    self._options = options or []
    self._description = description
    self._phase_dependencies = phase_dependencies or []
    self._add_default_options()

  def _add_default_options(self):
    return

  @property
  def phase_dependencies(self):
    return self._phase_dependencies

  @property
  def name(self):
    return self._name

  @property
  def options(self):
    return self._options

  def get_args(self, config=None):
    chain = itertools.chain.from_iterable(arg.get(config) for arg in self._args)
    return list(chain)

  @property
  def sources(self):
    return self._sources



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import re
import subprocess
import sys
import time

from twitter.common import log
from twitter.common.dirutil import safe_open

from twitter.pants import binary_util, get_buildroot
from twitter.pants.java import NailgunClient
from twitter.pants.tasks import Task


def _check_pid(pid):
  try:
    os.kill(pid, 0)
    return True
  except OSError:
    return False


class NailgunTask(Task):
  # Used to identify we own a given java nailgun server
  PANTS_NG_ARG_PREFIX = '-Dtwitter.pants.buildroot'
  PANTS_NG_ARG = '%s=%s' % (PANTS_NG_ARG_PREFIX, get_buildroot())

  _DAEMON_OPTION_PRESENT = False

  @staticmethod
  def create_pidfile_arg(pidfile):
    return '-Dpidfile=%s' % os.path.relpath(pidfile, get_buildroot())

  @staticmethod
  def _log_kill(log, pid, port=None):
    log.info('killing ng server @ pid:%d%s' % (pid, ' port:%d' % port if port else ''))

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    if not NailgunTask._DAEMON_OPTION_PRESENT:
      option_group.parser.add_option("--ng-daemons", "--no-ng-daemons", dest="nailgun_daemon",
                                     default=True, action="callback", callback=mkflag.set_bool,
                                     help="[%default] Use nailgun daemons to execute java tasks.")
      NailgunTask._DAEMON_OPTION_PRESENT = True

  def __init__(self, context, use_daemon=None, classpath=None, workdir=None, nailgun_args=None,
               stdin=None, stderr=sys.stderr, stdout=sys.stdout):
    Task.__init__(self, context)

    self._classpath = classpath
    self._nailgun_profile = context.config.get('nailgun', 'profile', default='nailgun')
    self._ng_server_args = nailgun_args or context.config.getlist('nailgun', 'args')
    self._stdin = stdin
    self._stderr = stderr
    self._stdout = stdout
    self._daemon = use_daemon if use_daemon is not None else context.options.nailgun_daemon

    workdir = workdir or os.path.join(context.config.get('nailgun', 'workdir'),
                                      self.__class__.__name__)
    self._pidfile = os.path.join(workdir, 'pid')
    self._ng_out = os.path.join(workdir, 'stdout')
    self._ng_err = os.path.join(workdir, 'stderr')

  def _runjava_common(self, main, classpath=None, opts=None, args=None, jvmargs=None):
    cp = (self._classpath or []) + (classpath or [])
    if self._daemon:
      nailgun = self._get_nailgun_client()
      try:
        if cp:
          nailgun('ng-cp', *[os.path.relpath(jar, get_buildroot()) for jar in cp])
        opts_args = []
        if opts:
          opts_args.extend(opts)
        if args:
          opts_args.extend(args)

        log.debug('Executing via nailgun %s: %s %s' % (nailgun, main, ' '.join(opts_args)))
        return nailgun(main, *opts_args)
      except nailgun.NailgunError as e:
        self._ng_shutdown()
        raise e
    else:
      return binary_util.runjava(main=main, classpath=cp, opts=opts, args=args, jvmargs=jvmargs)

  def runjava(self, main, classpath=None, opts=None, args=None, jvmargs=None):
    """Runs the java main using the given classpath and args.

    If --no-ng-daemons is specified then the java main is run in a freshly spawned subprocess,
    otherwise a persistent nailgun server dedicated to this Task subclass is used to speed up
    amortized run times. The args list is divisable so it can be split across multiple invocations
    of the command similiar to xargs.
    """

    return self._runjava_common(main, classpath=classpath, opts=opts, args=args, jvmargs=jvmargs)

  def runjava_indivisible(self, main, classpath=None, opts=None, args=None, jvmargs=None):
    """Runs the java main using the given classpath and args.

    If --no-ng-daemons is specified then the java main is run in a freshly spawned subprocess,
    otherwise a persistent nailgun server dedicated to this Task subclass is used to speed up
    amortized run times. The args list is indivisable so it can't be split across multiple
    invocations of the command similiar to xargs.
    """

    return self._runjava_common(main, classpath=classpath, opts=opts, args=args, jvmargs=jvmargs)

  def profile_classpath(self, profile):
    """Ensures the classpath for the given profile ivy.xml is available and returns it as a list of
    paths.

    If the classpath has changed since the last check for this profile this Task's build cache is
    invalidated.

    profile: The name of the tool profile classpath to ensure.
    """
    updated, classpath = binary_util.profile_classpath(profile,
                                                       java_runner=self.runjava_indivisible,
                                                       config=self.context.config)
    if updated:
      self.invalidate()
    return classpath

  def _ng_shutdown(self):
    endpoint = self._get_nailgun_endpoint()
    if endpoint:
      pid, port = endpoint
      NailgunTask._log_kill(self.context.log, pid, port)
      try:
        os.kill(pid, 9)
      except OSError:
        pass
      finally:
        os.remove(self._pidfile)

  def _get_nailgun_endpoint(self):
    if os.path.exists(self._pidfile):
      with safe_open(self._pidfile, 'r') as pidfile:
        contents = pidfile.read()

        def invalid_pidfile():
          log.warn('Invalid ng pidfile %s contained: %s' % (self._pidfile, contents))
          return None
        endpoint = contents.split(':')
        if len(endpoint) != 2:
          return invalid_pidfile()
        pid, port = endpoint
        try:
          return int(pid.strip()), int(port.strip())
        except ValueError:
          return invalid_pidfile()
    elif NailgunTask._find:
      pid_port = NailgunTask._find(self._pidfile)
      if pid_port:
        self.context.log.info('found ng server @ pid:%d port:%d' % pid_port)
        with safe_open(self._pidfile, 'w') as pidfile:
          pidfile.write('%d:%d\n' % pid_port)
      return pid_port
    return None

  def _get_nailgun_client(self):
    updated, ng_classpath = binary_util.profile_classpath(self._nailgun_profile)
    endpoint = self._get_nailgun_endpoint()
    running = _check_pid(endpoint[0]) if endpoint else False
    if running and not updated:
      return self._create_ngclient(port=endpoint[1])
    else:
      if running and updated:
        self._ng_shutdown()
      return self._spawn_nailgun_server(ng_classpath)

  # 'NGServer started on 127.0.0.1, port 53785.'
  _PARSE_NG_PORT = re.compile('.*\s+port\s+(\d+)\.$')

  def _parse_nailgun_port(self, line):
    match = NailgunTask._PARSE_NG_PORT.match(line)
    if not match:
      raise NailgunClient.NailgunError('Failed to determine spawned ng port from response'
                                       ' line: %s' % line)
    return int(match.group(1))

  def _await_nailgun_server(self):
    nailgun_timeout_seconds = 5
    max_socket_connect_attempts = 10
    nailgun = None
    port_parse_start = time.time()
    with safe_open(self._ng_out, 'r') as ng_out:
      while not nailgun:
        started = ng_out.readline()
        if started:
          port = self._parse_nailgun_port(started)
          with open(self._pidfile, 'a') as pidfile:
            pidfile.write(':%d\n' % port)
          nailgun = self._create_ngclient(port)
          log.debug('Detected ng server up on port %d' % port)
        elif time.time() - port_parse_start > nailgun_timeout_seconds:
          raise NailgunClient.NailgunError('Failed to read ng output after'
                                           ' %s seconds' % nailgun_timeout_seconds)

    attempt = 0
    while nailgun:
      sock = nailgun.try_connect()
      if sock:
        sock.close()
        log.info('Connected to ng server pid: %d @ port: %d' % self._get_nailgun_endpoint())
        return nailgun
      elif attempt > max_socket_connect_attempts:
        raise nailgun.NailgunError('Failed to connect to ng output after %d connect attempts'
                                   % max_socket_connect_attempts)
      attempt += 1
      log.debug('Failed to connect on attempt %d' % attempt)
      time.sleep(0.1)

  def _create_ngclient(self, port):
    return NailgunClient(port=port, work_dir=get_buildroot(), ins=self._stdin, out=self._stdout,
                         err=self._stderr)

  def _spawn_nailgun_server(self, ng_classpath):
    log.info('No ng server found, spawning...')

    with safe_open(self._ng_out, 'w'):
      pass  # truncate

    pid = os.fork()
    if pid != 0:
      # In the parent tine - block on ng being up for connections
      return self._await_nailgun_server()

    os.setsid()
    in_fd = open('/dev/null', 'w')
    out_fd = safe_open(self._ng_out, 'w')
    err_fd = safe_open(self._ng_err, 'w')

    args = ['java']
    if self._ng_server_args:
      args.extend(self._ng_server_args)
    args.append(NailgunTask.PANTS_NG_ARG)
    args.append(NailgunTask.create_pidfile_arg(self._pidfile))
    args.extend(['-cp', os.pathsep.join(ng_classpath),
                 'com.martiansoftware.nailgun.NGServer', ':0'])
    log.debug('Executing: %s' % ' '.join(args))

    with binary_util.safe_classpath(logger=log.warn):
      process = subprocess.Popen(
        args,
        stdin=in_fd,
        stdout=out_fd,
        stderr=err_fd,
        close_fds=True,
        cwd=get_buildroot()
      )
      with safe_open(self._pidfile, 'w') as pidfile:
        pidfile.write('%d' % process.pid)
      log.debug('Spawned ng server @ %d' % process.pid)
      # Prevents finally blocks and atexit handlers from being executed, unlike sys.exit(). We
      # don't want to execute finally blocks because we might, e.g., clean up tempfiles that the
      # parent still needs.
      os._exit(0)


try:
  import psutil

  def _find_ngs(everywhere=False):
    def cmdline_matches(cmdline):
      if everywhere:
        return any(filter(lambda arg: arg.startswith(NailgunTask.PANTS_NG_ARG_PREFIX), cmdline))
      else:
        return NailgunTask.PANTS_NG_ARG in cmdline

    for proc in psutil.process_iter():
      try:
        if 'java' == proc.name and cmdline_matches(proc.cmdline):
          yield proc
      except (psutil.AccessDenied, psutil.NoSuchProcess):
        pass

  def killall(log, everywhere=False):
    for proc in _find_ngs(everywhere=everywhere):
      try:
        NailgunTask._log_kill(log, proc.pid)
        proc.kill()
      except (psutil.AccessDenied, psutil.NoSuchProcess):
        pass

  NailgunTask.killall = staticmethod(killall)

  def _find_ng_listen_port(proc):
    for connection in proc.get_connections(kind='tcp'):
      if connection.status == 'LISTEN':
        host, port = connection.local_address
        return port
    return None

  def _find(pidfile):
    pidfile_arg = NailgunTask.create_pidfile_arg(pidfile)
    for proc in _find_ngs(everywhere=False):
      try:
        if pidfile_arg in proc.cmdline:
          port = _find_ng_listen_port(proc)
          if port:
            return proc.pid, port
      except (psutil.AccessDenied, psutil.NoSuchProcess):
        pass
    return None

  NailgunTask._find = staticmethod(_find)
except ImportError:
  NailgunTask.killall = None
  NailgunTask._find = None



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from functools import partial

from twitter.common.collections import maybe_list, OrderedSet

from twitter.pants.base import manual, Target
from twitter.pants.base.target import TargetDefinitionException
from twitter.pants.targets.anonymous import AnonymousDeps
from twitter.pants.targets.external_dependency import ExternalDependency


@manual.builddict(tags=["anylang"])
class JarLibrary(Target):
  """Serves as a proxy for one or more JarDependencies or JavaTargets."""

  def __init__(self, name, dependencies):
    """
    :param string name: The name of this module target, addressable via pants via the portion
      of the spec following the colon
    :param dependencies: one or more ``dependencies``, library targets,
      binary targets, or ``pants``.
    """

    Target.__init__(self, name)

    if dependencies is None:
      raise TargetDefinitionException(self, "dependencies are required")
    self.dependencies = OrderedSet(maybe_list(dependencies,
      expected_type=(ExternalDependency, AnonymousDeps, Target),
      raise_type=partial(TargetDefinitionException, self)))

  def resolve(self):
    yield self
    for dependency in self.dependencies:
      for resolved_dependency in dependency.resolve():
        yield resolved_dependency



from collections import defaultdict

from twitter.common.collections import OrderedDict, OrderedSet
from twitter.pants import is_internal
from twitter.pants.targets import InternalTarget
from twitter.pants.tasks import TaskError

class Group(object):
  @staticmethod
  def execute(phase, tasks_by_goal, context, executed, timer):
    """Executes the named phase against the current context tracking goal executions in executed."""

    def execute_task(name, task, targets):
      """Execute and time a single goal that has had all of its dependencies satisfied."""
      start = timer.now()
      try:
        # TODO (Senthil Kumaran):
        # Possible refactoring of the Task Execution Logic (AWESOME-1019)
        if getattr(context.options, 'explain', None):
          context.log.debug("Skipping execution of %s in explain mode" % name)
        else:
          task.execute(targets)
      finally:
        elapsed = timer.now() - start
        if phase not in executed:
          executed[phase] = OrderedDict()
        if elapsed:
          phase_timings = executed[phase]
          if name not in phase_timings:
            phase_timings[name] = []
          phase_timings[name].append(elapsed)

    tasks_by_goalname = dict((goal.name, task.__class__.__name__)
        for goal, task in tasks_by_goal.items())

    def expand_goal(goal):
      if len(goal) == 2: # goal is (group, goal)
        group_name, goal_name = goal
        task_name = tasks_by_goalname[goal_name]
        return "%s:%s->%s" % (group_name, goal_name, task_name)
      else:
        task_name = tasks_by_goalname[goal]
        return "%s->%s" % (goal, task_name)

    if phase not in executed:
      # Satisfy dependencies first
      goals = phase.goals()
      if not goals:
        raise TaskError('No goals installed for phase %s' % phase)

      for goal in goals:
        for dependency in goal.dependencies:
          Group.execute(dependency, tasks_by_goal, context, executed, timer)

      runqueue = []
      goals_by_group = {}
      for goal in goals:
        if goal.group:
          group_name = goal.group.name
          if group_name not in goals_by_group:
            group_goals = [goal]
            runqueue.append((group_name, group_goals))
            goals_by_group[group_name] = group_goals
          else:
            goals_by_group[group_name].append(goal)
        else:
          runqueue.append((None, [goal]))

      # OrderedSet takes care of not repeating chunked task execution mentions
      execution_phases = defaultdict(OrderedSet)

      for group_name, goals in runqueue:
        if not group_name:
          goal = goals[0]
          context.log.info('[%s:%s]' % (phase, goal.name))
          execution_phases[phase].add(goal.name)
          execute_task(goal.name, tasks_by_goal[goal], context.targets())
        else:
          for chunk in Group.create_chunks(context, goals):
            for goal in goals:
              goal_chunk = filter(goal.group.predicate, chunk)
              if len(goal_chunk) > 0:
                context.log.info('[%s:%s:%s]' % (phase, group_name, goal.name))
                execution_phases[phase].add((group_name, goal.name))
                execute_task(goal.name, tasks_by_goal[goal], goal_chunk)

      if getattr(context.options, 'explain', None):
        for phase, goals in execution_phases.items():
          goal_to_task = ", ".join(expand_goal(goal) for goal in goals)
          print("%s [%s]" % (phase, goal_to_task))

  @staticmethod
  def create_chunks(context, goals):
    def discriminator(target):
      for i, goal in enumerate(goals, start=1):
        if goal.group.predicate(target):
          return i
      return None

    # TODO(John Sirois): coalescing should be made available in another spot, InternalTarget is jvm
    # specific, and all we care is that the Targets have dependencies defined
    coalesced = InternalTarget.coalesce_targets(context.targets(is_internal), discriminator)
    coalesced = list(reversed(coalesced))

    def not_internal(target):
      return not is_internal(target)
    rest = OrderedSet(context.targets(not_internal))

    chunks = [rest] if rest else []
    flavor = None
    chunk_start = 0
    for i, target in enumerate(coalesced):
      target_flavor = discriminator(target)
      if target_flavor != flavor and i > chunk_start:
        chunks.append(OrderedSet(coalesced[chunk_start:i]))
        chunk_start = i
      flavor = target_flavor
    if chunk_start < len(coalesced):
      chunks.append(OrderedSet(coalesced[chunk_start:]))

    context.log.debug('::: created chunks(%d)' % len(chunks))
    for i, chunk in enumerate(chunks):
      context.log.debug('  chunk(%d):\n\t%s' % (i, '\n\t'.join(sorted(map(str, chunk)))))

    return chunks

  def __init__(self, name, predicate):
    self.name = name
    self.predicate = predicate



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = "tdesai"

import os
import re


class ArgError(Exception):
  """Raised to indicate a given Argument value is not valid."""


class BaseArg(object):
  def get(self, config=None):
    """
    Return an iterable of strings representing this argument
    """
    raise NotImplementedError("get must be implemented for all 'BaseArg's")


class Arg(BaseArg):
  """
  Key value arguments
  """
  """
  This Regular expression supports config value as /tmp/%(config.pants_workdir)s/thrift
  group1 is prefix or '/tmp'
  group2 is '%('
  group3 is value 'config.' or empty string if section is not present
  group4 is section value 'config' or empty string
  group5 is '.' if present or empty string
  group6 is option value 'pants_workdir'
  group7 is ')s'
  group 8 is suffix 's/thrift'
  """
  CONFIG_PATTERN = re.compile(r'(.*)(\%\()((.*)(\.))?(.+)(\)s)(.*)')

  def __init__(self, name, value, separator=None):
    self._name = name
    self._value = value
    self._separator = separator
    self._calculated_value = None

  def get(self, config):
    """
    >>> Arg('hello', 'world').get(None)
    ('hello', 'world')

    >>> Arg('-f', 'plain').get(None)
    ('-f', 'plain')

    >>> Arg('-k', '2', separator='=').get(None)
    ('-k=2',)
    """
    self._set_calculated_value(config)
    return ((self._name, self._calculated_value) if self._separator is None else
            (self._separator.join((self._name, self._calculated_value)),))

  def _set_calculated_value(self, config):
    if not self._calculated_value:
      self._calculated_value = self._interpolate_config(config, self._value)

  def _interpolate_config(self, config, value):
    config_key = Arg.CONFIG_PATTERN.match(value)
    if config_key:
      config_value = self._get_config_value(config, config_key.group(4), config_key.group(6))
      value = ''.join([config_key.group(1), config_value, config_key.group(8)])
      return self._interpolate_config(config, value)
    else:
      return value

  def _get_config_value(self, config, section, option):
    if not config:
      raise ArgError('Can not apply None config to value %s' % self._value)
    config_value = config.get(section, option) if section else config.getdefault(option)
    if not config_value:
      raise ArgError('Config value %s not set in the ini. Please check and re-run' % self._value)
    return config_value

  def get_value(self):
    return self._value


class FileArg(Arg):
  """
  Arguments whose value is a file
  """
  def __init__(self, name, file_name, separator=None, must_exist=False):
    super(FileArg, self).__init__(name, file_name, separator)
    self._must_exist = must_exist

  def file_exists(self):
    return os.path.exists(self._calculated_value)

  def get(self, config):
    """
    >>> FileArg('foo', 'bar', must_exist=False).get(None)
    ('foo', 'bar')

    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile() as temp_file:
    ...   print FileArg('-c', temp_file.name, must_exist=True).get(None) == ('-c', temp_file.name)
    True

    >>> FileArg('-foo', 'test', must_exist=True).get(None)
    Traceback (most recent call last):
    ...
    ArgError: File test does not exists for argument -foo
    """
    args = super(FileArg, self).get(config)
    if self._must_exist and not self.file_exists():
      raise ArgError("File %s does not exists for argument %s" % (self._calculated_value,
                                                                  self._name))
    return args


class JvmProperty(Arg):
  """
  Java property style arguments
  Generates a -D<key>=<value> flag
  """

  def __init__(self, key, value):
    self._key = key
    self._value = value
    self._calculated_value = None

  def get(self, config):
    """
    >>> JvmProperty('foo', 'bar').get(None)
    ('-Dfoo=bar',)

    >>> JvmProperty('foo', '').get(None)
    ('-Dfoo=',)
    """
    self._set_calculated_value(config)
    return ('-D%s=%s' % (self._key, self._calculated_value),)


class Flag(BaseArg):

  def __init__(self, flag):
    self._flag = flag

  def get(self, config):
    """
    >>> Flag('-foo').get(None)
    ('-foo',)

    >>> Flag('-usejavacp').get(None)
    ('-usejavacp',)
    """
    return ("%s" % (self._flag),)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import division, print_function

import errno
import glob
import hashlib
import os
import posixpath
import subprocess

from contextlib import closing, contextmanager

from twitter.common import log
from twitter.common.contextutil import environment_as, temporary_file
from twitter.common.dirutil import chmod_plus_x, safe_delete, safe_mkdir, safe_open, touch
from twitter.common.lang import Compatibility

if Compatibility.PY3:
  import urllib.request as urllib_request
  import urllib.error as urllib_error
else:
  import urllib2 as urllib_request
  import urllib2 as urllib_error

from twitter.pants.base import Config
from . import TaskError


_ID_BY_OS = {
  'linux': lambda release, machine: ('linux', machine),
  'darwin': lambda release, machine: ('darwin', release.split('.')[0]),
}


_PATH_BY_ID = {
  ('linux', 'x86_64'):  ['linux', 'x86_64'],
  ('linux', 'amd64'):   ['linux', 'x86_64'],
  ('linux', 'i386'):    ['linux', 'i386'],
  ('darwin', '9'):      ['mac', '10.5'],
  ('darwin', '10'):     ['mac', '10.6'],
  ('darwin', '11'):     ['mac', '10.7'],
  ('darwin', '12'):     ['mac', '10.8'],
  ('darwin', '13'):     ['mac', '10.9'],
}


def select_binary(base_path, version, name, config=None):
  """Selects a binary matching the current os and architecture.

  Raises TaskError if no binary of the given version and name could be found.
  """
  # TODO(John Sirois): finish doc of the path structure expexcted under base_path
  config = config or Config.load()
  cachedir = config.getdefault('pants_cachedir', default=os.path.expanduser('~/.pants.d'))
  baseurl = config.getdefault('pants_support_baseurl')
  timeout_secs = config.getdefault('pants_support_fetch_timeout_secs', type=int, default=30)

  sysname, _, release, _, machine = os.uname()
  os_id = _ID_BY_OS[sysname.lower()]
  if os_id:
    middle_path = _PATH_BY_ID[os_id(release, machine)]
    if middle_path:
      binary_path = os.path.join(base_path, *(middle_path + [version, name]))
      cached_binary_path = os.path.join(cachedir, binary_path)
      if not os.path.exists(cached_binary_path):
        url = posixpath.join(baseurl, binary_path)
        log.info('Fetching %s binary from: %s' % (name, url))
        downloadpath = cached_binary_path + '~'
        try:
          with closing(urllib_request.urlopen(url, timeout=timeout_secs)) as binary:
            with safe_open(downloadpath, 'wb') as cached_binary:
              cached_binary.write(binary.read())

          os.rename(downloadpath, cached_binary_path)
          chmod_plus_x(cached_binary_path)
        except (IOError, urllib_error.HTTPError, urllib_error.URLError) as e:
          raise TaskError('Failed to fetch binary from %s: %s' % (url, e))
        finally:
          safe_delete(downloadpath)
      log.debug('Selected %s binary cached at: %s' % (name, cached_binary_path))
      return cached_binary_path
  raise TaskError('No %s binary found for: %s' % (name, (sysname, release, machine)))


@contextmanager
def safe_args(args,
              max_args=None,
              config=None,
              argfile=None,
              delimiter='\n',
              quoter=None,
              delete=True):
  """
    Yields args if there are less than a limit otherwise writes args to an argfile and yields an
    argument list with one argument formed from the path of the argfile.

    :args The args to work with.
    :max_args The maximum number of args to let though without writing an argfile.  If not specified
              then the maximum will be loaded from config.
    :config Used to lookup the configured maximum number of args that can be passed to a subprocess;
            defaults to the default config and looks for key 'max_subprocess_args' in the DEFAULTS.
    :argfile The file to write args to when there are too many; defaults to a temporary file.
    :delimiter The delimiter to insert between args written to the argfile, defaults to '\n'
    :quoter A function that can take the argfile path and return a single argument value;
            defaults to:
            <code>lambda f: '@' + f<code>
    :delete If True deletes any arg files created upon exit from this context; defaults to True.
  """
  max_args = max_args or (config or Config.load()).getdefault('max_subprocess_args', int, 10)
  if len(args) > max_args:
    def create_argfile(fp):
      fp.write(delimiter.join(args))
      fp.close()
      return [quoter(fp.name) if quoter else '@%s' % fp.name]

    if argfile:
      try:
        with safe_open(argfile, 'w') as fp:
          yield create_argfile(fp)
      finally:
        if delete and os.path.exists(argfile):
          os.unlink(argfile)
    else:
      with temporary_file(cleanup=delete) as fp:
        yield create_argfile(fp)
  else:
    yield args


@contextmanager
def safe_classpath(logger=None):
  """
    Yields to a block in an environment with no CLASSPATH.  This is useful to ensure hermetic java
    invocations.
  """
  classpath = os.getenv('CLASSPATH')
  if classpath:
    logger = logger or log.warn
    logger('Scrubbing CLASSPATH=%s' % classpath)
  with environment_as(CLASSPATH=None):
    yield


def _runjava_cmd(jvmargs=None, classpath=None, main=None, opts=None, args=None):
  assert main or 1 == len(classpath), (
    "If 'main' is absent/None then 'classpath' must have one and only one element.")
  cmd = ['java']
  if jvmargs:
    cmd.extend(jvmargs)
  if classpath:
    cmd.extend(('-cp' if main else '-jar', os.pathsep.join(classpath)))
  if main:
    cmd.append(main)
  if opts:
    cmd.extend(opts)
  if args:
    cmd.extend(args)
  return cmd


def runjava_indivisible(jvmargs=None, classpath=None, main=None, opts=None, args=None, **kwargs):
  """Spawns a java process with the supplied configuration and returns its exit code.
  The args list is indivisable so it can't be split across multiple invocations of the command
  similiar to xargs.
  Passes kwargs through to subproccess.call.
  """
  cmd_with_args = _runjava_cmd(jvmargs=jvmargs, classpath=classpath, main=main, opts=opts,
                               args=args)
  with safe_classpath():
    return _subprocess_call(cmd_with_args, **kwargs)


def runjava(jvmargs=None, classpath=None, main=None, opts=None, args=None, **kwargs):
  """Spawns a java process with the supplied configuration and returns its exit code.
  The args list is divisable so it can be split across multiple invocations of the command
  similiar to xargs.
  Passes kwargs through to subproccess.call.
  """
  cmd = _runjava_cmd(jvmargs=jvmargs, classpath=classpath, main=main, opts=opts)
  with safe_classpath():
    return _subprocess_call_with_args(cmd, args, **kwargs)


def _split_args(i):
  l = list(i)
  half = len(l) // 2
  return l[:half], l[half:]


def _subprocess_call(cmd_with_args, call=subprocess.call, **kwargs):
  log.debug('Executing: %s' % ' '.join(cmd_with_args))
  return call(cmd_with_args, **kwargs)


def _subprocess_call_with_args(cmd, args, call=subprocess.call, **kwargs):
  cmd_with_args = cmd[:]
  if args:
    cmd_with_args.extend(args)
  try:
    with safe_classpath():
      return _subprocess_call(cmd_with_args, call=call, **kwargs)
  except OSError as e:
    if errno.E2BIG == e.errno and args and len(args) > 1:
      args1, args2 = _split_args(args)
      result1 = _subprocess_call_with_args(cmd, args1, call=call, **kwargs)
      result2 = _subprocess_call_with_args(cmd, args2, call=call, **kwargs)
      # we are making one command into two so if either fails we return fail
      result = 0
      if 0 != result1 or 0 != result2:
        result = 1
      return result
    else:
      raise e


def profile_classpath(profile, java_runner=None, config=None, ivy_jar=None, ivy_settings=None):
  """Loads the given profile's classpath if necessary and returns a tuple of
  (updated: bool, classpath: [string]).

  :param string profile: Name of a jvm tool profile.
  :param java_runner: Optional java command runner that accepts the following named parameters:
    jvmargs: An optional list of jvm arguments
    classpath: An optional list of classpath entries
    main: The optional name of of the main entrypoint
    opts: Options to pass to the main entrypoint
    args: Arguments to pass to the main entrypoint
  :param config: An optional site config.
  :type config: :class:`twitter.pants.base.Config`
  :param string ivy_jar: Optional path to a self-contained ivy jar.
  :param string ivy_settings: Optional path to an ivysettings.xml file.
  """
  # TODO(John Sirois): consider rework when ant backend is gone and there is no more need to share
  # path structure

  config = config or Config.load()
  profile_dir = config.get('ivy-profiles', 'workdir')
  ivy_xml = os.path.join(profile_dir, '%s.ivy.xml' % profile)
  if not os.path.exists(ivy_xml):
    raise TaskError('The ivy.xml to configure the %s tool classpath at %s '
                    'is missing.' % (profile, ivy_xml))

  digest = hashlib.sha1()
  with open(ivy_xml) as fp:
    digest.update(fp.read())

  profile_libdir = os.path.join(profile_dir, '%s.libs' % profile)
  profile_check = '%s.%s.checked' % (profile_libdir, digest.hexdigest())
  updated = False
  if not os.path.exists(profile_check):
    safe_mkdir(profile_libdir, clean=True)
    for path in glob.glob('%s*.checked' % profile_libdir):
      safe_delete(path)

    java_runner = java_runner or runjava_indivisible

    # TODO(John Sirois): refactor IvyResolve to share ivy invocation command line bits
    ivy_classpath = [ivy_jar] if ivy_jar else config.getlist('ivy', 'classpath')
    ivy_settings = ivy_settings or os.getenv('IVY_SETTINGS_XML') or config.get('ivy', 'ivy_settings')
    ivy_opts = [
      '-settings', ivy_settings,
      '-ivy', ivy_xml,

      # TODO(John Sirois): this pattern omits an [organisation]- prefix to satisfy IDEA jar naming
      # needs for scala - isolate this hack to idea.py where it belongs
      '-retrieve', '%s/[artifact]-[revision](-[classifier]).[ext]' % profile_libdir,

      # TODO(John Sirois): just use -cachepath and let ivy generate the classpath file contents
      # directly

      '-symlink',
      '-types', 'jar', 'bundle',
      '-confs', 'default'
    ]
    main = 'org.apache.ivy.Main'
    result = java_runner(classpath=ivy_classpath, main=main, opts=ivy_opts)
    if result != 0:
      raise TaskError("java %s ... exited non-zero (%i)"
                      " 'failed to load profile %s'" % (main, result, profile))
    with open(profile_check, 'w') as cp:
      for jar in os.listdir(profile_libdir):
        cp.write('%s\n' % os.path.join(profile_libdir, jar))
    updated = True

  with open(profile_check) as cp:
    return updated, map(lambda entry: entry.strip(), cp.readlines())


def get_tool_classpath(tool_lib_dir):
  return [os.path.join(tool_lib_dir, jar) for jar in os.listdir(tool_lib_dir)]


def _mac_open(files):
  subprocess.call(['open'] + list(files))


def _linux_open(files):
  for f in list(files):
    subprocess.call(['xdg-open', f])


_OPENER_BY_OS = {
  'darwin': _mac_open,
  'linux': _linux_open
}


def ui_open(*files):
  """Attempts to open the given files using the preferred native viewer or editor."""
  if files:
    osname = os.uname()[0].lower()
    if not osname in _OPENER_BY_OS:
      print('Sorry, open currently not supported for ' + osname)
    else:
      _OPENER_BY_OS[osname](files)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from .jvm_target import JvmTarget


class ExportableJvmLibrary(JvmTarget):
  """A baseclass for java targets that support being exported to an artifact repository."""

  def __init__(self,
               name,
               sources,
               provides=None,
               dependencies=None,
               excludes=None,
               buildflags=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param sources: A list of filenames representing the source code
      this library is compiled from.
    :type sources: list of strings
    :param provides:
      An optional Dependency object indicating the The ivy artifact to export.
    :param dependencies: List of :class:`twitter.pants.base.target.Target` instances
      this target depends on.
    :type dependencies: list of targets
    :param excludes: List of :class:`twitter.pants.targets.exclude.Exclude` instances
      to filter this target's transitive dependencies against.
    :param buildflags: Unused, and will be removed in a future release.
    """

    # it's critical provides is set 1st since _provides() is called elsewhere in the constructor
    # flow
    self._provides = provides

    JvmTarget.__init__(self, name, sources, dependencies, excludes, buildflags)

  @property
  def provides(self):
    return self._provides



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.common.dirutil import Fileset
from twitter.common.lang import Compatibility
from twitter.pants import is_concrete
from twitter.pants.base import manual, TargetDefinitionException

from .internal import InternalTarget
from .jvm_target import JvmTarget
from .resources import Resources


@manual.builddict(tags=["jvm"])
class JvmBinary(JvmTarget):
  """Produces a JVM binary optionally identifying a launcher main class.

  Invoking the ``binary`` goal on one of these targets creates a binary jar;
  the ``run`` goal executes it.

  If a ``main`` is specified, the binary can collect the main class from an
  depended-upon library source file, from a transitive dependency, or from a
  source specified in the ``source`` parameter.
  """
  def __init__(self, name,
               main=None,
               basename=None,
               source=None,
               resources=None,
               dependencies=None,
               excludes=None,
               deploy_excludes=None,
               configurations=None):
    """
    :param string name: The name of this target, which combined with this
      build file defines the target :class:`twitter.pants.base.address.Address`.
    :param string main: The name of the ``main`` class, e.g.,
      ``'com.twitter.common.examples.pingpong.Main'``.
    :param string basename: Base name for the generated ``.jar`` file, e.g.,
      ``'pingpong'``. (By default, uses ``name`` param)
    :param string source: Name of one ``.java`` or ``.scala`` file (a good
      place for a ``main``).
    :param resources: List of ``resource``\s to include in bundle.
    :param dependencies: List of targets (probably ``java_library`` and
     ``scala_library`` targets) to "link" in.
    :param excludes: List of ``exclude``\s to filter this target's transitive
     dependencies against.
    :param deploy_excludes: List of ``excludes`` to apply at deploy time.
      If you, for example, deploy a java servlet that has one version of
      ``servlet.jar`` onto a Tomcat environment that provides another version,
      they might conflict. ``deploy_excludes`` gives you a way to build your
      code but exclude the conflicting ``jar`` when deploying.
    :param configurations: Ivy configurations to resolve for this target.
      This parameter is not intended for general use.
    :type configurations: tuple of strings
    """
    JvmTarget.__init__(self,
                       name=name,
                       sources=[source] if source else None,
                       dependencies=dependencies,
                       excludes=excludes,
                       configurations=configurations)

    if main and not isinstance(main, Compatibility.string):
      raise TargetDefinitionException(self, 'main must be a fully qualified classname')

    if source and not isinstance(source, Compatibility.string):
      raise TargetDefinitionException(self, 'source must be a single relative file path')

    self.main = main
    self.basename = basename or name
    self.resources = list(self.resolve_all(resources, Resources))
    self.deploy_excludes = deploy_excludes or []


class RelativeToMapper(object):
  """A mapper that maps files specified relative to a base directory."""

  def __init__(self, base):
    """The base directory files should be mapped from."""

    self.base = base

  def __call__(self, file):
    return os.path.relpath(file, self.base)

  def __repr__(self):
    return 'IdentityMapper(%s)' % self.base


@manual.builddict(tags=["jvm"])
class Bundle(object):
  """Defines a bundle of files mapped from their full path name to a path name in the bundle."""

  def __init__(self, mapper=None, relative_to=None):
    """Creates a new bundle with an empty filemap.

    To add files, call the result's ``add`` method, passing in ``globs``,
    ``rglobs``, or paths. E.g., ``bundles = [ bundle().add(rglobs('config/*', 'scripts/*')) ]``.

    :param mapper: Function that takes a path string and returns a path string. Takes a path in
      the source tree, returns a path to use in the resulting bundle. By default, an identity mapper.
    :param string relative_to: Set up a simple mapping from source path to bundle path.
      E.g., ``bundle(relative_to='foo').add(rglobs('foo/bin/*'))`` to have those files show up
      in the bundle under ``bin`` instead of ``foo/bin``.
    """
    if mapper and relative_to:
      raise ValueError("Must specify exactly one of 'mapper' or 'relative_to'")

    if relative_to:
      base = os.path.abspath(relative_to)
      if not os.path.isdir(base):
        raise ValueError('Could not find a directory to bundle relative to at %s' % base)
      self.mapper = RelativeToMapper(base)
    else:
      self.mapper = mapper or RelativeToMapper(os.getcwd())

    self.filemap = {}

  @manual.builddict()
  def add(self, *filesets):
    for fileset in filesets:
      paths = fileset() if isinstance(fileset, Fileset) \
                        else fileset if hasattr(fileset, '__iter__') \
                        else [fileset]
      self.filemap.update(((os.path.abspath(path), self.mapper(path)) for path in paths))
    return self

  def resolve(self):
    yield self

  def __repr__(self):
    return 'Bundle(%s, %s)' % (self.mapper, self.filemap)


@manual.builddict(tags=["jvm"])
class JvmApp(InternalTarget):
  """Defines a jvm app package consisting of a binary plus additional bundles of files.

  Useful if you're building a service that you'll deploy to some other machine;
  that is, you're not just building a runnable binary, but also its data files.
  """

  def __init__(self, name, binary, bundles, basename=None):
    InternalTarget.__init__(self, name, dependencies=[binary])

    if not binary:
      raise TargetDefinitionException(self, 'binary is required')
    if not bundles:
      raise TargetDefinitionException(self, 'bundles must be specified')

    self._binary = binary
    self._bundles = bundles
    self.basename = basename or name

    self._resolved_binary = None
    self._resolved_bundles = []

  @property
  def binary(self):
    self._maybe_resolve_binary()
    return self._resolved_binary

  def _maybe_resolve_binary(self):
    if self._binary is not None:
      binaries = list(filter(is_concrete, self._binary.resolve()))
      if len(binaries) != 1 or not isinstance(binaries[0], JvmBinary):
        raise TargetDefinitionException(self,
                                        'must supply exactly 1 JvmBinary, got %s' % self._binary)
      self._resolved_binary = binaries[0]
      self._binary = None

  @property
  def bundles(self):
    self._maybe_resolve_bundles()
    return self._resolved_bundles

  def _maybe_resolve_bundles(self):
    if self._bundles is not None:
      def is_resolvable(item):
        return hasattr(item, 'resolve')

      def is_bundle(bundle):
        return isinstance(bundle, Bundle)

      def resolve(item):
        return list(item.resolve()) if is_resolvable(item) else [None]

      if is_resolvable(self._bundles):
        self._bundles = resolve(self._bundles)

      try:
        for item in iter(self._bundles):
          for bundle in resolve(item):
            if not is_bundle(bundle):
              raise TypeError()
            self._resolved_bundles.append(bundle)
      except TypeError:
        raise TargetDefinitionException(self, 'bundles must be one or more Bundle objects, '
                                              'got %s' % self._bundles)
      self._bundles = None






# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

__author__ = "tdesai"

from twitter.pants.goal import Goal
from twitter.pants.tasks.jvm_tool_task import JvmToolTask
from twitter.pants.tasks.ivy_resolver import IvyResolver


class JvmToolGoal(Goal):
  """
  JvmToolGoal resolves the jvm tool by calling make_runnable in prepare phase.
  """

  def __init__(self, name, action, jvm_tool_spec, group=None, dependencies=None):
    super(JvmToolGoal, self).__init__(name, action, group, dependencies)
    self.jvm_tool_spec = jvm_tool_spec

  def _make_runnable(self, context):
    IvyResolver(context, self.jvm_tool_spec).make_runnable()

  def prepare(self, context):
    self._make_runnable(context)
    return JvmToolTask(context=context, jvm_tool_spec=self.jvm_tool_spec)

  def task_setup_parser(self, group, args, mkflag):
    self._task.setup_parser(option_group=group, args=args, mkflag=mkflag,
                            options=self.jvm_tool_spec.options)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.common.dirutil import safe_open
from twitter.pants import has_sources, is_codegen
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.nailgun_task import NailgunTask


CHECKSTYLE_MAIN = 'com.puppycrawl.tools.checkstyle.Main'


class Checkstyle(NailgunTask):
  @staticmethod
  def _is_checked(target):
    return has_sources(target, '.java') and not is_codegen(target)

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    NailgunTask.setup_parser(option_group, args, mkflag)

    option_group.add_option(mkflag("skip"), mkflag("skip", negate=True),
                            dest="checkstyle_skip", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Skip checkstyle.")

  def __init__(self, context):
    self._profile = context.config.get('checkstyle', 'profile')
    workdir = context.config.get('checkstyle', 'nailgun_dir')
    NailgunTask.__init__(self, context, workdir=workdir)

    self._configuration_file = context.config.get('checkstyle', 'configuration')

    self._work_dir = context.config.get('checkstyle', 'workdir')
    self._properties = context.config.getdict('checkstyle', 'properties')
    self._confs = context.config.getlist('checkstyle', 'confs')

  def execute(self, targets):
    if not self.context.options.checkstyle_skip:
      checkstyle_classpath = self.profile_classpath(self._profile)
      with self.changed(filter(Checkstyle._is_checked, targets)) as changed_targets:
        sources = self.calculate_sources(changed_targets)
        if sources:
          result = self.checkstyle(checkstyle_classpath, sources)
          if result != 0:
            raise TaskError('java %s ... exited non-zero (%i)' % (CHECKSTYLE_MAIN, result))

  def calculate_sources(self, targets):
    sources = set()
    for target in targets:
      sources.update([os.path.join(target.target_base, source)
                      for source in target.sources if source.endswith('.java')])
    return sources

  def checkstyle(self, checkstyle_classpath, sources):
    with self.context.state('classpath', []) as cp:
      checkstyle_classpath.extend(jar for conf, jar in cp if conf in self._confs)

    opts = [
      '-c', self._configuration_file,
      '-f', 'plain'
    ]

    if self._properties:
      properties_file = os.path.join(self._work_dir, 'checkstyle.properties')
      with safe_open(properties_file, 'w') as pf:
        for k, v in self._properties.items():
          pf.write('%s=%s\n' % (k, v))
      opts.extend(['-p', properties_file])

    return self.runjava(CHECKSTYLE_MAIN, classpath=checkstyle_classpath, opts=opts, args=sources)
